// $Id: ListenOnur.java,v 1.10 2007/11/01 20:59:57 osoysal Exp $

/** modified from original Listen app. 
This one is specific to transact
*/
import java.io.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;
import net.tinyos.message.*;
import java.awt.*;
import java.awt.geom.*;
import javax.swing.*;
import java.util.*;

class UserInterface extends JFrame implements Runnable
{
	long lastUpdate;
	long initTime;
	JLabel lblTime;
	int nTasks;
	int nResources;
	int T_DELAY;
	int reportCount;
	int resourceReportCount;
	boolean requestDone;
	JTextArea taLog;
	JScrollPane jsp;
	public UserInterface(int T_DELAY,int nTasks,int nResources)
	{
		this.T_DELAY=T_DELAY;
		this.nTasks=nTasks;
		this.nResources=nResources;
		reportCount=0;
		resourceReportCount=0;
		requestDone=false;


		setBounds(new Rectangle(0,0,800,600));
		this.setLayout(null);
		this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		
		lblTime=new JLabel();
		lblTime.setBounds(new Rectangle(0,540,750,40));
		this.add(lblTime);
		
		taLog=new JTextArea();
		jsp=new JScrollPane(taLog);
		jsp.setBounds(new Rectangle(0,0,750,530));
		this.add(jsp);
		
		initTime=System.nanoTime();
		update();
		
		if (!ListenOnur.hideDisplay)
			this.setVisible(true);
	}
	public void run()
	{
		ListenOnur.println("Hello From Thread");
		byte []pkg=null;
		int iter;
		int MAX_DURATION=60; // in seconds
		try
		{
			for (iter=0;iter<MAX_DURATION*10;iter++) // run for approximately a minute
			{
				Thread.sleep(100);
				if (!requestDone && System.nanoTime()-lastUpdate>T_DELAY*100000000l)
				{
					for (int i=0;i<nTasks;i++)
					{
						ListenOnur.println("Requesting report from TASK "+i);
						pkg=ListenOnur.preparePacket(ListenOnur.prepareReportRequestContent((byte)i));
						ListenOnur.reader.writePacket(pkg);
						Thread.sleep(50);
					}

					for (int i=0;i<nResources;i++)
					{
						ListenOnur.println("Requesting report from RESOURCE "+i);
						pkg=ListenOnur.preparePacket(ListenOnur.prepareResourceReportRequestContent((byte)(i+nTasks)));
						ListenOnur.reader.writePacket(pkg);
						Thread.sleep(50);
					}
					
					ListenOnur.println("Requests done");
					requestDone=true;
					update(); // not exactly true but let it be 
				}
				if (reportCount == nTasks && resourceReportCount == nResources)
				{
					ListenOnur.println("Thread is dying gracefully");
					break;
				}
				updateLabel();
			}
			if (iter==MAX_DURATION*10)
			{
				reportCount = nTasks;
				ListenOnur.println("This has taken too long");
				if (!ListenOnur.keepOpen)
					System.exit(0);
			}
		}catch(Exception e)
		{
		}
	}
	
	public void updateLabel()
	{
		lblTime.setText("Current Time="+(System.nanoTime()-initTime)/1000000+"ms "+"Obtained results="+reportCount+"/"+nTasks);
	}
	public void update()
	{
		lastUpdate=System.nanoTime();
		updateLabel();
//		taLog.append("I've got a new update\r\n");
	}
	
	public void addToLog(String str)
	{
		taLog.append(str);
	}
	
	public void updateLogPos()
	{
		JScrollBar js;
		js=jsp.getVerticalScrollBar();
		js.setValue(js.getMaximum()-js.getVisibleAmount());
		taLog.setCaretPosition(taLog.getText().length());
	}
	
	public long currentTime()
	{
		return (System.nanoTime()-initTime)/1000000;
	}
	
}

public class ListenOnur 
{
	static String PacketType[]={"TI","TW","RR","CF","CM","CC","WA"};
	static String conflictNames [] = {	
		"CONFLICT_REASON_DETECTED",
		"CONFLICT_REASON_WRITE_MISSING",
		"CONFLICT_REASON_SOMEONE_DETECTED",
		"CONFLICT_REASON_RR_MISSING",
		"CONFLICT_REASON_RESPONSE_NOT_SENT",
		"CONFLICT_REASON_WRITE_ACK_MISSING", 
		"CONFLICT_REASON_COMPUTATION_NOT_READY",
		"CONFLICT_REASON_WRITE_ACK_NOT_SENT", 
		"CONFLICT_REASON_WRITE_NOT_SENT", 
		"CONFLICT_REASON_OTHER"
		};
	static int conflictCodes [] = {	42,	43,	44,	45,	46,	47,	48,	49,	50,	90 };
	
	static String getConflictName(int code)
	{
		for (int i=0;i<conflictCodes.length;i++)
			if (code == conflictCodes[i])
				return conflictNames[i];
		return "I_really_don't_know_this_one";
	}
	
	static PacketSource reader;
	static Random rnd;
	static UserInterface ui;
	static boolean keepOpen=false;
	static boolean briefDisplay=false;
	static boolean hideDisplay=false;
	static byte [][] readSets;
	static byte [][] writeSets;
	static boolean isVerified=true;
	static long firstPacketTime;
	static boolean firstPacketReceived=false;
	static long lastPacketTime;
	
	public static void println(String str)
	{
		System.out.println(str);
		ui.addToLog(str+"\n");
		ui.updateLogPos();
	}
	public static void println()
	{
		System.out.println();
		ui.addToLog("\n");
		ui.updateLogPos();
	}
	
	public static void print(String str)
	{
		System.out.print(str);
		System.out.flush();
		ui.addToLog(str);
	}
	
	public static void print(char c)
	{
		print(""+c);
	}
	
	public static void handleReport(byte [] packet)
	{
		int type=packet[10];
		int sender=packet[11];
		long timediff=((long)packet[12]&0xff)|(((long)packet[13]&0xff)<<8)|(((long)packet[14]&0xff)<<16)|(((long)packet[15]&0xff)<<24);
		int retries=((int)packet[16])|(((int)packet[17])<<8);
		println("Transaction on "+sender+" completed in "+timediff/1000.0+"ms. "+timediff+" with "+retries+" retransmissions");
	}
	
	public static int getTheoryValue(int resourceIdx)
	{
		int value=0;
		for (int i=0;i<writeSets.length;i++)
		{
			for (int j=0;j<writeSets[i].length;j++)
				if (writeSets[i][j]==resourceIdx)
					value=value+(1<<i);
		}
		return value;
	}

	public static void handleResourceReport(byte [] packet)
	{
		int type=packet[10];
		int sender=packet[11];
		int N_VARS=packet[12];
		byte [] values=new byte[N_VARS];
		print("Resource "+sender+" has final values=");
		for (int i=0;i<N_VARS;i++)
		{
			values[i]=packet[13+i];
			print(""+values[i]+",");
		}
		print(" Expected value was="+getTheoryValue(sender));
		println();
		if (values[0]!=getTheoryValue(sender))
			isVerified=false;
	}
	
	public static void prettyPrintReport(byte [] packet)
	{
		if (packet[10]==(byte)10)
			handleReport(packet);
		else if (packet[10]==(byte)11)
			handleResourceReport(packet);
		else
			println("DOHH!!! Unknown report");
	}
	
	public static void printPacket(byte [] packet)
	{
		for (int i=0;i<packet.length;i++)
		{
			int digit2=(packet[i]&0xFF)&0xF;
			int digit1=((packet[i]&0xFF)>>4)&0xF;
			if (digit1<10)
				print((char)('0'+digit1));
			else
				print((char)('A'+(digit1-10)));
			if (digit2<10)
				print((char)('0'+digit2));
			else
				print((char)('A'+(digit2-10)));
			print(' ');
		}
	}
	
	public static void prettyPrint(byte [] packet)
	{
		if (packet[8]==(byte)0x05)
		{
			prettyPrintReport(packet); // this is ugly!!!!
			return;
		}
		else if (packet[8]!=(byte)0x04) 
		{
			println("not a TRANSACT package ");
			return;
		}
		String str="";
		switch(packet[10])
		{
			case 0x10: // TI
				// nitems, (id,varid)*
				for (int i=0;i<packet[14];i++)
					str+="("+packet[15+i*2]+","+packet[15+i*2+1]+"),";
			break;
			case 0x11: // TW
				// nitems, (id, varid, value)*
				for (int i=0;i<packet[14];i++)
					str+="("+packet[15+i*3]+","+packet[15+i*3+1]+","+packet[15+i*3+2]+"),";
			break;
			case 0x12: // RR
				// responderid, value*
				str+=packet[14]+":";// responder
				for (int i=0;i<packet.length-15;i++) // actually listing these requires handling transactions
					str+=packet[15+i]+",";
			break;
			case 0x13: // CF
				// detectorID, conflict id [2bytes]
				str+=packet[14]+":";// detector id
				str+=packet[15]; // reason
				str+="["+getConflictName(packet[15])+"]";
			break;
			case 0x14: // CM
				str+=packet[14]+":";
				str+="["+getConflictName(packet[14])+"]";
				if (getConflictName(packet[14]).equals("CONFLICT_REASON_RR_MISSING"))
				{
					int count=packet[15];
					str+="={";
					for (int i=0;i<count;i++)
						str+=packet[16+i]+",";
					str+="}";
				}
				if (getConflictName(packet[14]).equals("CONFLICT_REASON_WRITE_ACK_MISSING"))
				{
					int count=packet[15];
					str+="={";
					for (int i=0;i<count;i++)
						str+=packet[16+i]+",";
					str+="}";
				}
			break;
			case 0x15: // CC
				str+=packet[14]+":"; // confirmer id
			break;
			case 0x16: // WA
				str+=packet[14]+":"; // confirmer id 
			break;
			default:
				str+="UNKNOWN packet";
			break;
		}
		print(""+ui.currentTime()+":Type: "+PacketType[packet[10]-0x10]);
		print(" Sender: "+packet[12]);
		print(" TID: "+packet[13]);
		println("   "+str);
	}
	

	// PACKET FORMAT
	// ***for assignment (type=0)
	// byte 0 -> type [0:assignment/1:start/2:reset]
	// byte 1 -> target mote
	// byte 2-> mote role [ 0 for task / 1 for resource] {not given in start}
	// byte 3 -> nread set {task only}
	// byte 4:4+nread-1  -> read set {task only}
	// byte 4+nread -> nwrite set {task only}
	// byte 4+nread+1:4+nread+1+nwrite-1 {task only}
	// ***for start (type=1)
	// byte 0 -> type [0:assignment/1:start]
	// byte 1 -> target mote
	// byte 2:3-> T_DELAY (Little endian)
	// byte 4:5-> T_INIT_BACK_OFF (Little endian)
	// byte 6 -> N_TRANSACTIONS
	// byte 7 -> REPEAT_FAILED [0 don't/ 1 do repeat]
	// ***for reset (type=2)
	// byte 0 -> type [0:assignment/1:start] 
	//***for report request(type=3)
	// byte 0 -> type 
	// byte 1 -> target mote
	// ***for resource report request(type = 4)
	// byte 0 -> type
	// byte 1 -> target mote
	// ***for report (type=10)
	// byte 0 -> type
	// byte 1 -> sender mote
	// byte 2:5 -> completion time in mili seconds (32 bit unsigned int)
	// byte 6 -> number of retrys
	// ***for resource report (type=11)
	// byte 0 -> type
	// byte 1 -> sender mote
	// byte 2 -> nTransactVars
	// byte 3:3+nTransactVars-1 -> currentValues

	public static byte[] prepareTaskContent(byte target, byte [] read, byte [] write)
	{
		byte [] content=new byte[5+read.length+write.length];
		content[0]=0; // assignment packet
		content[1]=target;
		content[2]=0; // task node
		content[3]=(byte)read.length;
		for (int i=0;i<read.length;i++)
			content[4+i]=read[i];
		content[4+read.length]=(byte)write.length;
		for (int i=0;i<write.length;i++)
			content[4+read.length+1+i]=write[i];
		return content;
	}

	public static byte[] prepareResourceContent(byte target)
	{
		byte [] content=new byte[3];
		content[0]=0; // assignment packet
		content[1]=target;
		content[2]=1; // resource node
		return content;
	}
	
	public static byte [] prepareStartContent(int T_DELAY,int N_TRANSACTIONS,int T_INIT_BACKOFF)
	{
		byte [] content=new byte[8];
		content[0]=1; // start packet
		content[1]=(byte)255; // everyone
		content[2]=(byte)(T_DELAY&255);
		content[3]=(byte)((T_DELAY>>8)&255);
		content[4]=(byte)(T_INIT_BACKOFF&255);
		content[5]=(byte)((T_INIT_BACKOFF>>8)&255);
		content[6]=(byte)N_TRANSACTIONS;
		content[7]=(byte)1; // assume always failures are repeated 
		return content;
	}
	
	public static byte [] prepareResetContent()
	{
		byte [] content=new byte[1];
		content[0]=2; // reset packet
		return content;
	}

	public static byte [] prepareReportRequestContent(byte target)
	{
		byte [] content=new byte[2];
		content[0]=3; // report packet
		content[1]=(byte)target;
		return content;
	}

	public static byte [] prepareResourceReportRequestContent(byte target)
	{
		byte [] content=new byte[2];
		content[0]=4; // resource report packet
		content[1]=(byte)target;
		return content;
	}
	
	public static byte[] preparePacket(byte [] content)
	{
		byte [] packet;
		byte[] header={(byte)0x03,(byte)0x02,(byte)0x08,(byte)0x9E,(byte)0xFF,(byte)0xFF,
						(byte)0xFF,(byte)0xFF,(byte)0x05,(byte)0x7D};
		packet=new byte[header.length+content.length];
		for (int i=0;i<header.length;i++)
			packet[i]=header[i];
		for (int i=0;i<content.length;i++)
			packet[header.length+i]=content[i];
		packet[0]=(byte)content.length;
		return packet;
	}
	
	public static byte [] getReadSet(int nTasks,int nResources)
	{
		boolean [] isIn=new boolean[nResources];
		byte [] read;
		int nRead=0;
		do
		{
			for (int i=0;i<nResources;i++)
			{
				isIn[i]=rnd.nextBoolean();
				if (isIn[i])
					nRead++;
			}
		}while(nRead==0);
		read=new byte[nRead];
		int j=0;
		for (int i=0;i<nResources;i++)
			if (isIn[i])
				read[j++]=(byte)(i+nTasks);
		return read;
	}
	
	public static byte [] getWriteSet(byte [] read)
	{
		byte [] write;
		boolean [] isIn=new boolean[read.length];
		int nWrite=0;
		do
		{
			for (int i=0;i<read.length;i++)
			{
				isIn[i]=rnd.nextBoolean();
				if (isIn[i])
					nWrite++;
			}
		}
		while (nWrite==0);
		write=new byte[nWrite];
		int j=0;
		for (int i=0;i<read.length;i++)
			if (isIn[i])
				write[j++]=(byte)read[i];
		return write;
	}
	
	public static void runExperiment(int T_DELAY,int nTasks,int nResources)
	{
		ui=new UserInterface(T_DELAY,nTasks,nResources);
		Thread t=new Thread(ui);
		t.start();
		byte [] read;
		byte [] write;
		byte []pkg=null;
		reader = BuildSource.makePacketSource();
		if (reader == null) {
		    System.err.println("Invalid packet source (check your MOTECOM environment variable)");
		    System.exit(2);
		}
		
		try {
			reader.open(PrintStreamMessenger.err);
			Thread.sleep(200);
			ui.update();
			print("Reseting nodes ...");
			System.out.flush();
			pkg=preparePacket(prepareResetContent());
			reader.writePacket(pkg);
			println("DONE");
			readSets=new byte[nTasks][];
			writeSets=new byte[nTasks][];
			for (int i=0;i<nTasks;i++)
			{
				print("Assigning to "+i+"  TASK  ");
				read=getReadSet(nTasks,nResources);
				write=getWriteSet(read);
				readSets[i]=read;
				writeSets[i]=write;
				print("Read={"+read[0]);
				for (int j=1;j<read.length;j++)
					print(","+read[j]);
				print("} "); 
				print("Write={"+write[0]);
				for (int j=1;j<write.length;j++)
					print(","+write[j]);
				print("} ..."); 
				System.out.flush();
				
				pkg=preparePacket(prepareTaskContent((byte)i,read,write));
				reader.writePacket(pkg);
				println("DONE");
				Thread.sleep(200);
				ui.update();
			}
			
			for (int i=0;i<nResources;i++)
			{
				print("Assigning to "+(nTasks+i)+"  RESOURCE ...");
				System.out.flush();
				pkg=preparePacket(prepareResourceContent((byte)(nTasks+i)));
				reader.writePacket(pkg);
				println("DONE");
				Thread.sleep(200);
				ui.update();
			}
			Thread.sleep(1000);
			ui.update();

			println("STARTING...");
			pkg=preparePacket(prepareStartContent(T_DELAY,1,0));
			reader.writePacket(pkg);

			for (int i=0;;i++) 
			{
				byte[] packet = reader.readPacket();
				if (packet[8]==(byte)0x04 )// transact packet
				{
					if (!firstPacketReceived)
					{
						firstPacketReceived=true;
						firstPacketTime=ui.currentTime();
					}
					lastPacketTime=ui.currentTime();
				}
				if (!briefDisplay)
				{
					printPacket(packet);
					println();
				}
				prettyPrint(packet);
				if (packet[8]==(byte)0x05 && packet[10]==(byte)10)
					ui.reportCount++;
				if (packet[8]==(byte)0x05 && packet[10]==(byte)11)
					ui.resourceReportCount++;
				ui.update();
				if (ui.reportCount==nTasks && ui.resourceReportCount==nResources)
				{
					println("Run completed!");
					println("Time between last and first packets = "+(lastPacketTime-firstPacketTime)/1000.0 +" sec");
					print("Result is ... ");
					if (isVerified)
						println("SUCCESS");
					else
						println("FAIL");
					reader.close();
					break;
				}
			}
		}
		catch (Exception e) 
		{
			println("Got an exception: "+e);
			e.printStackTrace();
			// thread will kill it closing by closing the descriptor
		}
		println("Thread Run Time="+ui.currentTime()/1000.0+" sec");
	}
	
	
    public static void main(String args[]) throws IOException 
	{
		if (args.length < 4) 
		{
			System.err.println("usage: java ListenOnur T_DELAY nTasks nResources seed [-k] [-b] [-s]");
			System.err.println("-k: keep window open after finishes");
			System.err.println("-b: just pretty print messages");
			System.err.println("-s: hides java display");
			System.exit(2);
		}
		int T_DELAY=Integer.parseInt(args[0]);
		int nTasks=Integer.parseInt(args[1]);
		int nResources=Integer.parseInt(args[2]);
		int seed=Integer.parseInt(args[3]);
		for (int i=4;i<args.length;i++)
		{
			if (args[i].equals("-k"))
				keepOpen=true;
			else if (args[i].equals("-b"))
				briefDisplay=true;
			else if (args[i].equals("-s"))
				hideDisplay=true;
		}		
		rnd=new Random(seed);
		runExperiment(T_DELAY,nTasks,nResources);
		if (!keepOpen)
			System.exit(0);
	}
}



