// $Id: TransactInterface.h,v 1.22 2007/11/02 19:35:21 osoysal Exp $
#ifndef TRANSACT_INTERFACE_H
#define TRANSACT_INTERFACE_H

#ifndef MAX_TRANSACTIONS
	#define MAX_TRANSACTIONS 10     // number of concurrent transaction allowed
#endif
#define N_TRANSACT_VARS 7
#define MAX_SAFE_TRANSACT_VARS 5

#ifndef T_DELAY_DEFAULT
	#define T_DELAY_DEFAULT 512
#endif

uint32_t T_DELAY=T_DELAY_DEFAULT;

#define T_READ_PENDING (T_DELAY)
#define T_WAIT_READ_RESPONSE (T_DELAY*3)
#define T_WAIT_CALCULATION (T_DELAY)
#define T_WRITE_PENDING (T_DELAY)
#define T_WAIT_CONFLICT (T_DELAY)
#define T_CANCEL_PENDING (T_DELAY)
#define T_WAIT_CONFIRM (T_DELAY)
#define T_WAIT_WRITE_ACK (T_DELAY*2)
#ifndef T_GRACE_PERIOD
// first two shouldn't be needed. 
	#define T_GRACE_PERIOD (T_WAIT_READ_RESPONSE+T_WAIT_CALCULATION+T_WRITE_PENDING+T_WAIT_WRITE_ACK+T_WAIT_CONFLICT+T_CANCEL_PENDING+T_WAIT_CONFIRM)
#endif

#define T_TRANSACTION_DURATION (T_READ_PENDING+T_WAIT_READ_RESPONSE+T_WAIT_CALCULATION+T_WRITE_PENDING+T_WAIT_CONFLICT+T_CANCEL_PENDING+T_WAIT_CONFIRM+T_WAIT_WRITE_ACK+T_GRACE_PERIOD)

/** Conflict Result codes */
enum
{
	CONFLICT_REASON_DETECTED  = 42,
	CONFLICT_REASON_WRITE_MISSING = 43,
	CONFLICT_REASON_SOMEONE_DETECTED = 44,
	CONFLICT_REASON_RR_MISSING = 45,
	CONFLICT_REASON_RESPONSE_NOT_SENT = 46,
	CONFLICT_REASON_WRITE_ACK_MISSING = 47,
	CONFLICT_REASON_COMPUTATION_NOT_READY = 48,
	CONFLICT_REASON_WRITE_ACK_NOT_SENT = 49,
	CONFLICT_REASON_WRITE_NOT_SENT = 50,
	CONFLICT_REASON_OTHER = 90
};

/** Type of transact variables */
typedef uint8_t TransactVarType;

/** each item in read list */
typedef struct 
{
	/** id of the node to be read */
	uint8_t nodeID;
	/** id of the varialbe */
	uint8_t varID;
}readListItem;

/** each item in write list */
typedef struct 
{
	/** id of the node to be updated */
	uint8_t nodeID;
	/** id of the variable to be updated */
	uint8_t varID;
	/** desired value of the variable */
	TransactVarType value;
}writeListItem;

/** Transaction initiation message. 
the fields must be properly filled 
*/
typedef struct 
{
	/** number of read operations */
	uint8_t nReadList; 
	/** the list of reads */
	readListItem readList[N_TRANSACT_VARS];
}TransactionInit;         

/**
Transaction write message 
the fields are similar to the read but also a value should be provided
*/
typedef struct
{
	/** number of write operations */
	uint8_t nWriteList;  
	/** the list of write operations */
	writeListItem writeList[N_TRANSACT_VARS];
}TransactionWrite;

/**
This structure contains the values read from other nodes in a read request
*/
typedef struct 
{		
	/** used to check if current transaction has received all the read responses */
	bool responseReceived[N_TRANSACT_VARS+2]; 
	/** each response entry corresponds to the readList item with the same index in the initiation message*/
	TransactVarType response[N_TRANSACT_VARS+2];  
	/** this flag is set if all the responses have been recieved */
	bool allDone;
}TransactionReadState;
/* -------INTERNAL TYPES ------*/

/** this message is sent for reporting conflicts */
typedef struct 
{
	uint8_t detectorID;
	uint8_t conflictType;
}ConflictMessage;

/** this message is sent to declare cancellation of a transaction */
typedef struct 
{
	uint8_t reason;// see the conflict reasons
	uint8_t count; // if readrespose or writeack is missing this is the number of missing ones
	uint8_t extraData[N_TRANSACT_VARS]; // and this contains list of the nodes responsible
}CancelMessage;

/** this message is sent to confirm cancellation of a message */
typedef struct 
{
	uint8_t confirmerID;
}CancelConfirmation;

/** this message is sent to confirm a write message */
typedef struct 
{
	uint8_t confirmerID;
}WriteAck;

/**
This structure is sent for responding to read requests. This message is intended for internal use. 
*/
typedef struct 
{
	/** the id of the responding node */
	uint8_t responderID;
	/** a list of values. Only indexes corresponding to this node in initiation message are valid */
	TransactVarType values[N_TRANSACT_VARS]; 
}ReadResponse;

/** this enumerates the types of transact messages */
typedef enum 
{
TI=0x10, ///< transaction initiation message
TW=0x11, ///< transaction write message
RR=0x12, ///< read response message
CF=0x13, ///< conflict detection message
CM=0x14, ///< cancel message
CC=0x15, ///< cancel confirmation message
WA=0x16  ///< write ack message
} Msg_type ;

/**
This structure is the real payload for Transact messages. It is intended for internal use only. 
*/
typedef struct 
{
	Msg_type type;
	uint8_t senderID;
	uint8_t transactionID;
	uint8_t data[21];
}TransactMessage;
#endif
